/* upsconf.h - prototypes for upsconf.c

   Copyright (C) 2001  Russell Kroll <rkroll@exploits.org>
	2005	Arnaud Quette <arnaud.quette@free.fr>
	2013	Emilien Kia <kiae.dev@gmail.com>
	2020-2026	Jim Klimov <jimklimov+nut@gmail.com>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#ifndef NUT_UPSCONF_H_SEEN
#define NUT_UPSCONF_H_SEEN 1

#ifdef __cplusplus
/* *INDENT-OFF* */
extern "C" {
/* *INDENT-ON* */
#endif

/* callback function from read_upsconf */
void do_upsconf_args(char *upsname, char *var, char *val);

/* pointer to it that dynamically-linked private NUT libraries can also call;
 * might be a bit of an overkill with legacy static linking.
 */
extern void (*callback_upsconf_args)(char *upsname, char *var, char *val);

/* open the ups.conf, parse it, and call back do_upsconf_args()
 * returns -1 (or aborts the program) in case of errors;
 * returns 1 if processing finished successfully
 * See also reload_flag support in main.c for live-reload feature
 */
int read_upsconf(int fatal_errors);

#ifdef __cplusplus
/* *INDENT-OFF* */
}
/* *INDENT-ON* */
#endif

#endif	/* NUT_UPSCONF_H_SEEN */
