initDefines() {
i=1
ERR_ROOT="$i";i="`expr $i + 1`"
ERR_UNKNOWN_SYS_STR1="$i";i="`expr $i + 1`"
ERR_UNKNOWN_SYS_STR2="$i";i="`expr $i + 1`"
ERR_UNKNOWN_SYS_STR3="$i";i="`expr $i + 1`"
CS_ERR_NO_SERIAL="$i";i="`expr $i + 1`"
CS_ERR_NO_NUM="$i";i="`expr $i + 1`"
WELCOME_STR1="$i";i="`expr $i + 1`"
WELCOME_STR2="$i";i="`expr $i + 1`"
WELCOME_STR3="$i";i="`expr $i + 1`"
INSTALL_INTRO_STR1="$i";i="`expr $i + 1`"
INSTALL_INTRO_STR2="$i";i="`expr $i + 1`"
INSTALL_INTRO_STR3="$i";i="`expr $i + 1`"
INSTALL_INTRO_STR4="$i";i="`expr $i + 1`"
STR_VERSION="$i";i="`expr $i + 1`"
DEFAULT_PATH_STR="$i";i="`expr $i + 1`"
LICENSE_AGREEMENT_1="$i";i="`expr $i + 1`"
LICENSE_AGREEMENT_2="$i";i="`expr $i + 1`"
LICENSE_AGREEMENT_3="$i";i="`expr $i + 1`"
LICENSE_AGREEMENT_4="$i";i="`expr $i + 1`"
LICENSE_AGREEMENT_5="$i";i="`expr $i + 1`"
ERR_LIC_AGREE="$i";i="`expr $i + 1`"
SUCCESS_INSTALL="$i";i="`expr $i + 1`"
UNSUCCESS_INSTALL="$i";i="`expr $i + 1`"
INSTALL_PACKAGE="$i";i="`expr $i + 1`"
REMOVE_PACKAGE="$i";i="`expr $i + 1`"
INSTALL_ERROR="$i";i="`expr $i + 1`"
CS_SEPARATOR_1="$i";i="`expr $i + 1`"
CS_WELCOME_1="$i";i="`expr $i + 1`"
# UPS server Configuration
CS_SERVER_1="$i";i="`expr $i + 1`"
CS_SERVER_2="$i";i="`expr $i + 1`"
CS_SERVER_3="$i";i="`expr $i + 1`"
CS_SERVER_4="$i";i="`expr $i + 1`"
CS_SERVER_5="$i";i="`expr $i + 1`"
CS_SERVER_6="$i";i="`expr $i + 1`"
CS_SERVER_7="$i";i="`expr $i + 1`"
CS_SERVER_8="$i";i="`expr $i + 1`"
# Select UPS Password
CS_PASSWORD_1="$i";i="`expr $i + 1`"
CS_PASSWORD_2="$i";i="`expr $i + 1`"
CS_PASSWORD_3="$i";i="`expr $i + 1`"
CS_PASSWORD_4="$i";i="`expr $i + 1`"
#Please review your configuration
CS_REVIEW_1="$i";i="`expr $i + 1`"
CS_REVIEW_2="$i";i="`expr $i + 1`"
CS_REVIEW_3="$i";i="`expr $i + 1`"
CS_REVIEW_4="$i";i="`expr $i + 1`"
CS_REVIEW_5="$i";i="`expr $i + 1`"
CS_REVIEW_6="$i";i="`expr $i + 1`"
CS_REVIEW_7="$i";i="`expr $i + 1`"
CS_REVIEW_8="$i";i="`expr $i + 1`"
CS_REVIEW_9="$i";i="`expr $i + 1`"
CS_REVIEW_10="$i";i="`expr $i + 1`"
CS_CONNECTIVITY_1="$i";i="`expr $i + 1`"
CS_CONNECTIVITY_2="$i";i="`expr $i + 1`"
CS_CONNECTIVITY_3="$i";i="`expr $i + 1`"
CS_CONNECTIVITY_4="$i";i="`expr $i + 1`"
CS_CONNECTIVITY_5="$i";i="`expr $i + 1`"
# Verify Detection
CS_ASK_SERIAL_1="$i";i="`expr $i + 1`"
CS_ASK_SERIAL_2="$i";i="`expr $i + 1`"
CS_ASK_SERIAL_3="$i";i="`expr $i + 1`"
CS_ASK_SERIAL_4="$i";i="`expr $i + 1`"
CS_ASK_SERIAL_5="$i";i="`expr $i + 1`"
CS_ASK_SERIAL_6="$i";i="`expr $i + 1`"
# Choose serial device
CS_SERIAL_1="$i";i="`expr $i + 1`"
CS_SERIAL_3="$i";i="`expr $i + 1`"
CS_SERIAL_4="$i";i="`expr $i + 1`"
# Manually choose serial device
CS_MANUAL_SERIAL_1="$i";i="`expr $i + 1`"
CS_MANUAL_SERIAL_2="$i";i="`expr $i + 1`"
CS_MANUAL_SERIAL_3="$i";i="`expr $i + 1`"
CS_MANUAL_SERIAL_ERR="$i";i="`expr $i + 1`"
# SNMP Configuration
CS_SNMP_1="$i";i="`expr $i + 1`"
CS_SNMP_2="$i";i="`expr $i + 1`"
CS_SNMP_3="$i";i="`expr $i + 1`"
CS_SNMP_4="$i";i="`expr $i + 1`"
CS_SNMP_5="$i";i="`expr $i + 1`"
CS_SNMP_6="$i";i="`expr $i + 1`"
CS_SNMP_7="$i";i="`expr $i + 1`"
CS_SNMP_8="$i";i="`expr $i + 1`"
#0. Go back
CS_SNMP_9="$i";i="`expr $i + 1`"
#Scan for more UPS
CS_SNMP_10="$i";i="`expr $i + 1`"
#Starting UPP service
CS_START_SERVICE="$i";i="`expr $i + 1`"
#LanSafe installation detected. Proceeding to uninstallation.
CS_REMOVE_LS="$i";i="`expr $i + 1`"
CS_REMOVE_NS="$i";i="`expr $i + 1`"
CS_REMOVE_OLDIPP="$i";i="`expr $i + 1`"
CS_SHUTOFF_ENABLE_1="$i";i="`expr $i + 1`"
CS_SHUTOFF_ENABLE_2="$i";i="`expr $i + 1`"
CS_SHUTOFF_DELAY_1="$i";i="`expr $i + 1`"
CS_SHUTOFF_DELAY_2="$i";i="`expr $i + 1`"
CS_SHUTOFF_DELAY_3="$i";i="`expr $i + 1`"
CS_MANAGE_1="$i";i="`expr $i + 1`"
CS_MANAGE_2="$i";i="`expr $i + 1`"
CS_MANAGE_3="$i";i="`expr $i + 1`"
CS_MANAGE_4="$i";i="`expr $i + 1`"
CS_NETWORK_1="$i";i="`expr $i + 1`"
CS_NETWORK_2="$i";i="`expr $i + 1`"
CS_NETWORK_3="$i";i="`expr $i + 1`"
CS_NETWORK_4="$i";i="`expr $i + 1`"
CS_NETWORK_5="$i";i="`expr $i + 1`"
CS_ANOTHER_UPS_1="$i";i="`expr $i + 1`"
CS_ANOTHER_UPS_2="$i";i="`expr $i + 1`"
CS_ANOTHER_UPS_3="$i";i="`expr $i + 1`"
CS_ANOTHER_UPS_4="$i";i="`expr $i + 1`"
CS_ANOTHER_UPS_5="$i";i="`expr $i + 1`"
CS_ANOTHER_UPS_6="$i";i="`expr $i + 1`"
CS_MINSUP_1="$i";i="`expr $i + 1`"
CS_MINSUP_2="$i";i="`expr $i + 1`"
CS_REMOVE_UPS_1="$i";i="`expr $i + 1`"
CS_REMOVE_UPS_2="$i";i="`expr $i + 1`"
CS_NMC_LOGIN_1="$i";i="`expr $i + 1`"
CS_NMC_LOGIN_2="$i";i="`expr $i + 1`"
CS_NMC_LOGIN_3="$i";i="`expr $i + 1`"
CS_NMC_LOGIN_4="$i";i="`expr $i + 1`"
CS_ERR_PASSWORDS_DIFFER="$i";i="`expr $i + 1`"
CS_CLIENT_SD_DURATION_1="$i";i="`expr $i + 1`"
CS_CLIENT_SD_DURATION_2="$i";i="`expr $i + 1`"
CS_XML_MANU_AUTO_1="$i";i="`expr $i + 1`"
CS_XML_MANU_AUTO_2="$i";i="`expr $i + 1`"
CS_XML_MANU_AUTO_3="$i";i="`expr $i + 1`"
CS_XML_MANU_AUTO_4="$i";i="`expr $i + 1`"
CS_XML_MANU_AUTO_5="$i";i="`expr $i + 1`"
CS_XML_MANU_AUTO_6="$i";i="`expr $i + 1`"
CS_XML_MANU_1="$i";i="`expr $i + 1`"
CS_XML_MANU_2="$i";i="`expr $i + 1`"
CS_STATIC_POWERCYCLE_1="$i";i="`expr $i + 1`"
CS_STATIC_POWERCYCLE_2="$i";i="`expr $i + 1`"
CS_ERR_BADSTRING="$i";i="`expr $i + 1`"
CS_ESDTIMER_1="$i";i="`expr $i + 1`"
CS_ESDTIMER_2="$i";i="`expr $i + 1`"
}

necho() {
        lineStr="`head -$1 $installres |tail -1`"
        echo "$lineStr""$2""$3""$4""$5""$6""$7""$8""$9"
}
