/*  nutbooltest.c - test custom nut_bool_t usability
 *
 *  Copyright (C)
 *      2024            Jim Klimov <jimklimov+nut@gmail.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include "config.h"
#include "nut_bool.h"

#include <stdio.h>
#include <stdlib.h>

int main(void)
{
	nut_bool_t bt = true, bf = false;
	int ret = 0;

	/* Check basic boolean operations */
	if (bf)		ret++;
	if (!bt)	ret++;
	if (!bt == !bf)	ret++;
	if (bt == bf)	ret++;

	if (!(!bt == bf))	ret++;
	if (!(bt == !bf))	ret++;

	if (!(bt != bf))	ret++;
	if (!(!bt != !bf))	ret++;

	if (bf && bt)		ret++;
	if (!bf && !bt)		ret++;

	if (!(!bf && bt))	ret++;
	if (!(bf || bt))	ret++;
	if (!(!bf || !bt))	ret++;

	if (ret != 0)
		printf("nutbooltest collected %i errors", ret);

	return (ret != 0);
}
